#include "../lib/mysql.hpp"

marcelb::mysql::MySQL::MySQL(const string _path, const string _username, const string _password, const string _db, const uint32_t _available):
#ifdef MYSQL_USE_ASYNCO
   p_loop(periodic( [&] () {
      const uint64_t start = rtime_ms();
      auto pairs = divide_and_conquer(connection_pool.size(), t);

      vector<future<void>> fuve;
      for (auto& pair : pairs) {
         fuve.push_back(async_ ([&, pair](){
            _tloop(pair.first, pair.second);
         }));
      }

      for (auto& fu : fuve) {
         try {
            await_(fu);
         } catch (...) {
            if(on_error) {
               on_error("Error in maintenance periodic loop.");
            }
         }
      }
      
      auto_adjusts(rtime_ms() - start);

   }, MYSQL_PERIODIC_INTERNAL_TIME)),
#else 
   tloop_future (async(launch::async, [&](){
      while (run_tloop) {
         usleep(MYSQL_PERIODIC_INTERNAL_TIME*1000);

         const uint64_t start = timestamp();
         auto pairs = divide_and_conquer(connection_pool.size(), t);

         vector<future<void>> fuve;
         for (auto& pair : pairs) {
            fuve.push_back(async (launch::async, [&, pair](){
               _tloop(pair.first, pair.second);
            }));
         }

         for (auto& fu : fuve) {
            try {
               fu.get();
            } catch (...) {
               if(on_error) {
                  on_error("Error in maintenance periodic loop.");
               }
            }
         }

         auto_adjusts(timestamp() - start);
      }
      return;
   })),
#endif
   path(_path),
   username(_username),
   password(_password),
   database(_db),
   pool_size(_available > 0 ? _available : 1) {
   
   drv = get_mysql_driver_instance();
   connect_pool();
}


Connection* marcelb::mysql::MySQL::create_connection() {
   uint32_t trys = 0;
   bool status = true;
   Connection* new_con = NULL;
   while (connect_trys == unlimited ? status : (trys <= connect_trys && status)) {
      try {
         Connection* con_can = drv->connect(path, username, password);
         con_can->setSchema(database);
         status = !con_can->isValid();
         if (!status) {
            new_con = con_can;
         }
         else if (!con_can->isClosed()) {
            disconnect_connection(con_can);
         }
      }
      catch (const SQLException &error) {
         if (on_error) {
            on_error(error.what() + string(", SQL state: ") + error.getSQLState() + string(", Error code: ") + to_string(error.getErrorCode()));
         }
         usleep(reconnectSleep);
         connect_trys == unlimited ? trys : trys++;
      }         
   }

   return new_con;
}

void marcelb::mysql::MySQL::connect_pool() {
   auto connect_ = [&]() -> void {
      const uint32_t tens = pool_size/MYSQL_MAX_CONNECTION_INIT_SAME_TIME;
      for (uint32_t i=0; i<pool_size; i++) {
         Connection* connection = create_connection();
         {  
            lock_guard<mutex> lock(io);
            connection_pool.push(connection);
         }
         if ((i+1)%tens == 0) {
            usleep(MYSQL_DELAYED_CONNECT_TIME*1000);
         }
      }
      if (on_connect) {
         on_connect();
      }
   };

#ifdef MYSQL_USE_ASYNCO
   async_ (connect_);
#else 
   connect_f = async (launch::async, connect_);
#endif
}

void marcelb::mysql::MySQL::disconnect_pool() {
   lock_guard<mutex> lock(io);
   for (uint32_t i=0; i<connection_pool.size(); i++) {
      Connection* connection = connection_pool.front();
      connection_pool.pop();
      disconnect_connection(connection) ;
   }
}

bool marcelb::mysql::MySQL::disconnect_connection(Connection* connection) {
   bool status = !connection->isClosed();
   if (status) {
      try {
         connection->close();
         status = !connection->isClosed();
      } 
      catch (const SQLException &error) {
         if (on_error) {
            on_error(error.what() + string(", SQL state: ") + error.getSQLState() + string(", Error code: ") + to_string(error.getErrorCode()));
         }
         status = true;
      }
   }

   else {
      status = false; // već je zatvorena
   }

   delete connection;
   return status;
}

void marcelb::mysql::MySQL::_tloop(uint32_t b, uint32_t e) {
   for (size_t i=b; i<connection_pool.size() && i<e; i++) {
      try {
         Connection *conn = nullptr;      // provjeri ovdje svugdje možeš li koristiti release i ocupacy sada
         conn = occupy_connection();
         if (conn->isValid()) {
            release_connection(conn);
         } else {
            if (!conn->isClosed()){
               disconnect_connection(conn);
            }
            Connection *n_conn = create_connection();
            release_connection(n_conn);

         }
      } catch (const SQLException &error) {
         if (on_error) {
            on_error(error.what() + string(", SQL state: ") + error.getSQLState() + string(", Error code: ") + to_string(error.getErrorCode()));
         }
      }
   }
}

Connection* marcelb::mysql::MySQL::occupy_connection() {
   unique_lock<mutex> lock(io);
   while (connection_pool.empty()) {
      condition.wait(lock);
   }
   Connection *connection = connection_pool.front();
   connection_pool.pop();
   return connection;
}

void marcelb::mysql::MySQL::release_connection(Connection* connection) {
   lock_guard<std::mutex> lock(io);
   connection_pool.push(connection);
   condition.notify_one();
}

marcelb::mysql::MySQL::~MySQL() {
#ifdef MYSQL_USE_ASYNCO
   p_loop.stop();
#else
   run_tloop = false;
   tloop_future.get();
#endif
   disconnect_pool();
}